import numpy as np
from scipy.optimize import golden
from scipy.linalg import norm, solve

class MinimalizacniMetoda:
    def __init__(self, f, df):
        self.f = f
        self.df = df
        return None

    def d(x):
        pass

    def krok(self, x):
        d = self.d(x)
        def fk(t):
            return self.f(x + t * d)
        alpha = golden(fk)
        return (x + alpha * d, self.f(x + alpha * d))        

class NejvetsiSpad(MinimalizacniMetoda):
    def d(self, x):
        return -self.df(x)

class SdruzeneGradienty(MinimalizacniMetoda):
    def __init__(self, f, df):
        MinimalizacniMetoda.__init__(self, f, df)
        self.ck = None # gradient v predchozim kroku
        self.dk = None # smer postupu v predchozim kroku
        return None

    def d(self, x):
        if self.ck is None:
            out = -self.df(x)
        else:
            out = -self.df(x) + (
                norm(self.df(x)) /
                norm(self.ck))**2 * self.dk
        self.ck = self.df(x)
        self.dk = out
        return out

class Newton(MinimalizacniMetoda):
    def __init__(self, f, df, H):
        MinimalizacniMetoda.__init__(self, f, df)
        self.H = H
        return None

    def d(self, x):
        return solve(-self.H(x), self.df(x))

class BFGS(MinimalizacniMetoda):
    def __init__(self, f, df, ndim=2):
        MinimalizacniMetoda.__init__(self, f, df)
        self.Bk = np.eye(ndim)
        self.xk = None
        self.ck = None
        return None

    def d(self, x):
        if self.xk is None:
            B = self.Bk
        else:
            s = x - self.xk
            y = self.df(x) - self.ck
            Bk = self.Bk
            B = Bk - np.dot(np.dot(Bk, s), np.dot(s, Bk)) \
                / np.dot(s.T, np.dot(Bk, s)) \
                + np.dot(y.T, y) / np.dot(s.T, s)
        return solve(B, self.df(x))

if __name__ == "__main__":
    N = 3 # pocet kroku u kazde metody
    H = np.array([[1., .5],[.5, .75]])
    def f(x):
        return np.dot(x.T, np.dot(H, x))

    def df(x):
        return 2 * np.dot(H, x)

    def res(x, f, metoda):
        for ii in range(N):
            xi, fi = metoda.krok(x[-1])
            x.append(xi)
            f.append(fi)
        return None

    def vypis_prubeh(xi, fi):
        for ii, xf in enumerate(zip(xi, fi)):
            print 'x%d = %f, y%d = %f, f%d = %f' % (ii, xf[0][0], ii, xf[0][1], ii, xf[1])
        return None

    print 'Metoda nejvetsiho spadu:'
    nejvetsi_spad = NejvetsiSpad(f, df)
    x_ns = [np.array([0,1.5]),]
    f_ns = [f(x_ns[0]),]

    res(x_ns, f_ns, nejvetsi_spad)
    vypis_prubeh(x_ns, f_ns)

    print 'Metoda sdruzenych gradientu:'
    sdruzene_grad = SdruzeneGradienty(f, df)
    x_sg = [np.array([0, 1.5]),]
    f_sg = [f(x_sg[0]),]

    res(x_sg, f_sg, sdruzene_grad)
    vypis_prubeh(x_sg, f_sg)

    print 'Newtonova metoda:'
    newton = Newton(f, df, lambda x: H)
    x_n = [np.array([0, 1.5]),]
    f_n = [f(x_sg[0]),]

    res(x_n, f_n, newton)
    vypis_prubeh(x_n, f_n)

    print 'BFGS:'
    bfgs = BFGS(f, df)

    x_b = [np.array([0, 1.5]),]
    f_b = [f(x_sg[0]),]

    res(x_b, f_b, newton)
    vypis_prubeh(x_b, f_b)

    x = np.linspace(-2, 2, 21)
    y = np.linspace(-2, 2, 21)
    import matplotlib.pyplot as plt
    plt.contour(x, y, np.array([[f(np.array([xi, yi])) for
                                 xi in x] for yi in y]))

    def vykresli(x, label, marker='o:'):
        plt.plot([xi[0] for xi in x],
                 [xi[1] for xi in x],
                 marker, label=label)

    vykresli(x_ns, 'Nejvetsi spad')
    vykresli(x_sg, 'Sdruzene gradienty', '^:')
    vykresli(x_n, 'Newtonova metoda', 'v:')
    vykresli(x_b, 'BFGS', '.:')

    plt.xlabel('$x$')
    plt.ylabel('$y$')
    plt.grid()
    plt.legend(loc='best')
    plt.show()
